from enum import IntEnum


class ReversibleIntEnum(IntEnum):
    @classmethod
    def as_dict(cls):
        return cls.__members__

    @classmethod
    def reverse_dict(cls):
        return {v: k for k, v in cls.__members__.items()}

    def to_idx(self):
        return self - 1



from dataclasses import dataclass

import numpy as np

from ..protocol.dataset_proto import Context, Transform
from ..protocol.label_proto import Label, Polygon2dProto


@dataclass
class SimplifiedFrame:
    context: Context
    timestamp_micros: int
    pose: Transform
    laser_labels: [Label]
    no_label_zones: [Polygon2dProto]
    points: np.ndarray



"""
Based on https://github.com/waymo-research/waymo-open-dataset/blob/master/waymo_open_dataset/protos/keypoint.proto
"""

from dataclasses import dataclass

from .utils import ReversibleIntEnum


@dataclass
class Vec2d:
    x: float
    y: float


@dataclass
class Vec3d:
    x: float
    y: float
    z: float


@dataclass
class KeypointVisibility:
    is_occluded: bool


@dataclass
class Keypoint2d:
    location_px: Vec2d
    visibility: KeypointVisibility


@dataclass
class Keypoint3d:
    visibility: KeypointVisibility
    location_px: Vec3d = None


class KeypointType(ReversibleIntEnum):
    KEYPOINT_TYPE_UNSPECIFIED = 0
    # Tip of nose.
    KEYPOINT_TYPE_NOSE = 1
    KEYPOINT_TYPE_LEFT_SHOULDER = 5
    KEYPOINT_TYPE_LEFT_ELBOW = 6
    KEYPOINT_TYPE_LEFT_WRIST = 7
    KEYPOINT_TYPE_LEFT_HIP = 8
    KEYPOINT_TYPE_LEFT_KNEE = 9
    KEYPOINT_TYPE_LEFT_ANKLE = 10
    KEYPOINT_TYPE_RIGHT_SHOULDER = 13
    KEYPOINT_TYPE_RIGHT_ELBOW = 14
    KEYPOINT_TYPE_RIGHT_WRIST = 15
    KEYPOINT_TYPE_RIGHT_HIP = 16
    KEYPOINT_TYPE_RIGHT_KNEE = 17
    KEYPOINT_TYPE_RIGHT_ANKLE = 18
    # Center of the forehead area.
    KEYPOINT_TYPE_FOREHEAD = 19
    # A point in the center of head - a point in the middle between two ears.
    # The nose and head center together create an imaginary line in the direction
    # that the person is looking (i.e. head orientation).
    KEYPOINT_TYPE_HEAD_CENTER = 20


@dataclass
class CameraKeypoint:
    type: KeypointType
    keypoint_2d: Keypoint2d = None
    keypoint_3d: Keypoint3d = None


@dataclass
class CameraKeypoints:
    keypoint: [CameraKeypoint]


@dataclass
class LaserKeypoint:
    type: KeypointType
    keypoint_3d: Keypoint3d


@dataclass
class LaserKeypoints:
    keypoint: [LaserKeypoint]



import argparse
import io
import math

from torch_waymo import WaymoDataset

try:
    from PIL import Image
except ImportError as e:
    raise ImportError("Pillow is required for this demo. Install with `pip install pillow`.") from e

import matplotlib.pyplot as plt
import numpy as np

from torch_waymo.protocol.dataset_proto import CameraName
from torch_waymo.protocol.label_proto import Type as LabelType


def _decode_image(data: bytes) -> np.ndarray:
    with Image.open(io.BytesIO(data)) as im:
        return np.array(im.convert("RGB"))


def _describe_frame(frame, index: int) -> str:
    lines = []
    frame_type = frame.__class__.__name__
    timestamp = getattr(frame, "timestamp_micros", "N/A")

    # Cameras
    if hasattr(frame, "images"):
        cameras = frame.images
        cam_names = []
        for ci in cameras:
            try:
                cam_names.append(CameraName(ci.name).name)
            except Exception:
                cam_names.append(str(ci.name))
        cam_summary = f"{len(cameras)} cameras: [" + ", ".join(cam_names) + "]"
    else:
        cam_summary = "0 cameras (simplified frame)"

    # Lasers
    if hasattr(frame, "lasers"):
        laser_count = len(frame.lasers)
    else:
        laser_count = 0

    # Labels
    labels = getattr(frame, "laser_labels", []) or []
    total_labels = len(labels)
    label_counts = {lt: 0 for lt in LabelType}
    for lbl in labels:
        try:
            label_counts[lbl.type] += 1
        except Exception:
            pass
    label_parts = []
    for lt, cnt in label_counts.items():
        if cnt > 0:
            label_parts.append(f"{lt.name.replace('TYPE_', '')}:{cnt}")
    label_summary = f"labels: {total_labels}" + (" (" + ", ".join(label_parts) + ")" if label_parts else "")

    # Points
    pts_obj = frame.points
    if pts_obj is None:
        point_summary = "points: N/A"
    else:
        if isinstance(pts_obj, (list, tuple)):
            total_pts = sum(getattr(p, "shape", [0])[0] if hasattr(p, "shape") else 0 for p in pts_obj)
        else:
            total_pts = getattr(pts_obj, "shape", [0])[0] if hasattr(pts_obj, "shape") else 0
        point_summary = f"points: {total_pts}"

    lines.append(f"Frame {index} | {frame_type}")
    lines.append(f" timestamp: {timestamp}")
    lines.append(f" {cam_summary}")
    lines.append(f" lasers: {laser_count}")
    lines.append(f" {label_summary}")
    lines.append(f" {point_summary}")
    return "\n" + "\n".join(lines) + "\n"


def _show_frames(dataset_path: str, split: str, num_frames: int):
    dataset = WaymoDataset(dataset_path, split)
    for i in range(num_frames):
        frame = dataset[i]
        print(_describe_frame(frame, i))
        if hasattr(frame, "images"):
            cam_images = frame.images
            if len(cam_images) == 0:
                print(f"Frame {i} has no images.")
            else:
                n = len(cam_images)
                cols = min(3, n)
                rows = math.ceil(n / cols)
                fig, axes = plt.subplots(rows, cols, figsize=(4 * cols, 3.5 * rows))
                if not isinstance(axes, np.ndarray):  # single axis
                    axes = np.array([[axes]])
                axes = axes.reshape(rows, cols)

                for idx, cam_img in enumerate(cam_images):
                    r = idx // cols
                    c = idx % cols
                    ax = axes[r, c]
                    data = cam_img.image
                    if isinstance(data, bytes):
                        try:
                            arr = _decode_image(data)
                        except Exception as e:
                            print(f"Decode failure for frame {i} image {idx}: {e}")
                            arr = np.zeros((32, 32, 3), dtype=np.uint8)
                    else:
                        arr = data
                        if arr.ndim == 2:
                            arr = np.repeat(arr[..., None], 3, axis=-1)
                        elif arr.shape[2] == 4:
                            arr = arr[..., :3]
                    # Title from enum name
                    try:
                        cam_name = CameraName(cam_img.name).name
                    except Exception:
                        cam_name = str(cam_img.name)
                    ax.imshow(arr)
                    ax.set_title(cam_name)
                    ax.axis("off")

                for extra_idx in range(n, rows * cols):
                    r = extra_idx // cols
                    c = extra_idx % cols
                    axes[r, c].axis("off")

                fig.suptitle(f"Frame {i} - All camera images ({n})")
                plt.tight_layout()
                plt.show()
        else:
            print(f"Frame {i} is a SimplifiedFrame (no images available).")


def main():
    parser = argparse.ArgumentParser(description="Show Waymo dataset camera images in a grid per frame.")
    parser.add_argument("-d", "--dataset", required=True, help="Path to converted or converted_simplified dataset root")
    parser.add_argument(
        "-s", "--split", choices=["training", "validation", "testing"], default="training", help="Dataset split"
    )
    parser.add_argument("-n", "--num-frames", type=int, default=10, help="Number of frames to display")
    args = parser.parse_args()

    _show_frames(args.dataset, args.split, args.num_frames)


if __name__ == "__main__":
    main()



"""
Based on https://github.com/waymo-research/waymo-open-dataset/blob/master/waymo_open_dataset/dataset.proto#L145
"""

import typing
from dataclasses import dataclass, fields, is_dataclass

import numpy as np

from .label_proto import Label, Polygon2dProto, Type
from .utils import ReversibleIntEnum

T = typing.TypeVar("T")


def fullname(o):
    """
    Get the full name of a class, including module, e.g. 'module.submodule.ClassName'
    :param o: object or class
    :return: full name
    """
    klass = o.__class__
    module = klass.__module__
    if module == "builtins":
        return klass.__qualname__  # avoid outputs like 'builtins.str'
    return module + "." + klass.__qualname__


def from_data(cls: typing.Type[T], data) -> T:
    """
    Convert a protobuf message to a dataclass instance.
    Handles nested dataclasses and lists of dataclasses.
    :param cls: the dataclass type to convert to
    :param data: the protobuf message
    :return: the dataclass instance
    """
    from google.protobuf.pyext._message import (
        RepeatedCompositeContainer,
        RepeatedScalarContainer,
    )
    from waymo_open_dataset import dataset_pb2

    if isinstance(data, dataset_pb2.Transform):
        return np.array(data.transform).reshape((4, 4))
    if (
        isinstance(data, list)
        or isinstance(data, RepeatedCompositeContainer)
        or isinstance(data, RepeatedScalarContainer)
    ):
        return [from_data(cls[0], d) for d in data]

    if not is_dataclass(cls):
        return data

    field_names = [f.name for f in fields(cls)]
    field_types = {f.name: f.type for f in fields(cls)}

    attributes = dict()
    for name in dir(data):
        if name in field_names:
            field_type = field_types[name]
            field_data = getattr(data, name)
            attributes[name] = from_data(field_type, field_data)

    return cls(**attributes)


Transform = np.ndarray


class CameraName(ReversibleIntEnum):
    """
    Camera names as per the Waymo dataset specification.
    """

    UNKNOWN = 0
    FRONT = 1
    FRONT_LEFT = 2
    FRONT_RIGHT = 3
    SIDE_LEFT = 4
    SIDE_RIGHT = 5


class LaserName(ReversibleIntEnum):
    """
    Laser names as per the Waymo dataset specification.
    """

    UNKNOWN = 0
    TOP = 1
    FRONT = 2
    SIDE_LEFT = 3
    SIDE_RIGHT = 4
    REAR = 5


@dataclass
class Velocity:
    """
    Velocity of the sensor in the vehicle frame.
    """

    # velocity in m/s
    v_x: float
    v_y: float
    v_z: float

    # angular velocity in rad/s
    w_x: float
    w_y: float
    w_z: float


class RollingShutterReadOutDirection(ReversibleIntEnum):
    """
    Rolling shutter readout direction as per the Waymo dataset specification.
    """

    UNKNOWN = 0
    TOP_TO_BOTTOM = 1
    LEFT_TO_RIGHT = 2
    BOTTOM_TO_TOP = 3
    RIGHT_TO_LEFT = 4
    GLOBAL_SHUTTER = 5


@dataclass
class CameraCalibration:
    """
    Camera calibration parameters.
    """

    name: CameraName
    # 1d Array of [f_u, f_v, c_u, c_v, k{1, 2}, p{1, 2}, k{3}], follows the same definition as OpenCV
    intrinsic: [float]
    extrinsic: Transform
    width: int
    height: int
    rolling_shutter_direction: RollingShutterReadOutDirection


@dataclass
class LaserCalibration:
    """
    Laser calibration parameters.
    """

    name: LaserName
    beam_inclinations: [float]
    beam_inclination_min: float
    beam_inclination_max: float
    extrinsic: Transform


@dataclass
class ObjectCount:
    """
    Object count for a specific type.
    """

    type: Type
    count: int


@dataclass
class Stats:
    """
    Dataset statistics for the context.
    """

    laser_object_counts: [ObjectCount]
    camera_object_counts: [ObjectCount]
    time_of_day: str
    location: str
    weather: str


@dataclass
class Context:
    """
    Context information for the sequence.
    """

    name: str
    camera_calibrations: [CameraCalibration]
    laser_calibrations: [LaserCalibration]
    stats: Stats


@dataclass
class RangeImage:
    """
    Range image and associated data.
    *_compressed fields are compressed using Zlib, decompress using:
    val = ZlibDecompress(range_image_compressed)
    """

    range_image_compressed: np.ndarray
    camera_projection_compressed: np.ndarray
    range_image_pose_compressed: np.ndarray
    range_image_flow_compressed: np.ndarray
    segmentation_label_compressed: np.ndarray


@dataclass
class InstanceIDToGlobalIDMapping:
    """
    Mapping from local instance ID to global instance ID.
    """

    local_instance_id: int
    global_instance_id: int
    is_tracked: bool


@dataclass
class CameraSegmentationLabel:
    """
    Camera segmentation label information.
    """

    panoptic_label_divisor: int
    panoptic_label: np.ndarray
    instance_id_to_global_id_mapping: [InstanceIDToGlobalIDMapping]
    sequence_id: str


@dataclass
class CameraImage:
    """
    Camera image and associated data.
    """

    name: CameraName
    image: bytes
    pose: Transform
    velocity: Velocity
    pose_timestamp: float
    shutter: float
    camera_trigger_time: float
    camera_readout_done_time: float
    camera_segmentation_label: CameraSegmentationLabel


@dataclass
class CameraLabels:
    """
    Labels associated with a camera.
    """

    name: CameraName
    labels: [Label]


@dataclass
class Laser:
    """
    Laser data and associated range images.
    """

    name: LaserName
    ri_return1: RangeImage
    ri_return2: RangeImage


@dataclass
class Frame:
    """
    A single frame of data in the Waymo dataset.
    """

    context: Context
    timestamp_micros: int
    pose: Transform
    images: [CameraImage]
    lasers: [Laser]
    laser_labels: [Label]
    projected_lidar_labels: [CameraLabels]
    camera_labels: [CameraLabels]
    no_label_zones: [Polygon2dProto]
    # Empty if not computed
    points: np.ndarray = None



"""
Based on https://github.com/waymo-research/waymo-open-dataset/blob/master/waymo_open_dataset/label.proto
"""

from dataclasses import dataclass

from .keypoint_proto import CameraKeypoints, LaserKeypoints
from .utils import ReversibleIntEnum


class BoxType(ReversibleIntEnum):
    TYPE_UNKNOWN = 0
    # 7-DOF 3D (a.k.a upright 3D box).
    TYPE_3D = 1
    # 5-DOF 2D. Mostly used for laser top down representation.
    TYPE_2D = 2
    # Axis aligned 2D. Mostly used for image.
    TYPE_AA_2D = 3


@dataclass
class Box:
    center_x: float
    center_y: float
    center_z: float

    length: float
    width: float
    height: float

    # angle in rad (-pi, pi(, rotate +x to the box front face
    heading: float


@dataclass
class Metadata:
    speed_x: float
    speed_y: float
    speed_z: float

    accel_x: float
    accel_y: float
    accel_z: float


class Type(ReversibleIntEnum):
    TYPE_UNKNOWN = 0
    TYPE_VEHICLE = 1
    TYPE_PEDESTRIAN = 2
    TYPE_SIGN = 3
    TYPE_CYCLIST = 4


class DifficultyLevel(ReversibleIntEnum):
    UNKNOWN = 0
    LEVEL_1 = 1
    LEVEL_2 = 2


@dataclass
class Association:
    laser_object_id: str


@dataclass
class Label:
    box: Box
    metadata: Metadata
    type: Type
    id: str
    detection_difficulty_level: DifficultyLevel
    tracking_difficulty_level: DifficultyLevel
    num_lidar_points_in_box: int
    num_top_lidar_points_in_box: int
    laser_keypoints: LaserKeypoints
    camera_keypoints: CameraKeypoints
    association: Association
    most_visible_camera_name: str
    camera_synced_box: Box


@dataclass
class Polygon2dProto:
    x: [float]
    y: [float]
    id: str



import pathlib
import pickle
from typing import Callable, Optional, Union

from torch.utils.data import Dataset

from ..protocol.dataset_proto import Frame
from .simplified_frame import SimplifiedFrame


class WaymoDataset(Dataset):
    def __init__(self, root_path: str, split: str, transform: Optional[Callable] = None):
        self._root_path = pathlib.Path(root_path).expanduser()
        self._split = split
        self._split_path = self._root_path.joinpath(split)
        self._transform = transform

        # Validate root path existence
        if not self._root_path.exists() or not self._root_path.is_dir():
            raise FileNotFoundError(f"Dataset root path does not exist or is not a directory: {self._root_path}")

        # Validate split directory existence
        if not self._split_path.exists() or not self._split_path.is_dir():
            raise FileNotFoundError(
                f"Split path does not exist: {self._split_path}. You may have the wrong root or forgot to convert this split."
            )

        self._seq_len_cache_path = self._split_path.joinpath("len.pkl")
        if self._seq_len_cache_path.exists():
            with open(self._seq_len_cache_path, "rb") as f:
                self._seq_lens = pickle.load(f)
        else:
            raise FileNotFoundError(
                f"Could not find sequence length cache file: {self._seq_len_cache_path}. Conversion may be incomplete."
            )

    def __len__(self) -> int:
        return sum(self._seq_lens)

    def __getitem__(self, idx: int) -> Union[SimplifiedFrame, Frame]:
        path = self._split_path.joinpath(f"{idx}.pkl")
        if path.exists():
            return self._get_frame(path)
        else:
            raise IndexError(f"Could not load frame at index {idx} (missing file {path}).")

    def _get_frame(self, path):
        with open(path, "rb") as f:
            x = pickle.load(f)
        if self._transform is not None:
            x = self._transform(x)
        return x



import argparse
import pathlib
import pickle

import tensorflow.compat.v1 as tf

tf.enable_eager_execution()
import tqdm
from waymo_open_dataset import dataset_pb2 as open_dataset
from waymo_open_dataset.utils import frame_utils

from torch_waymo.dataset import SimplifiedFrame
from torch_waymo.protocol import dataset_proto
from torch_waymo.protocol.dataset_proto import Frame


def generate_cache(root_path: pathlib.Path, split: str, simplified: bool = False):
    """
    Convert the Waymo Open Dataset to a format without Tensorflow dependencies.
    :param root_path: Path to the Waymo dataset root directory.
    :param split: One of "training", "validation", "testing".
    :param simplified: If True, store simplified frames (no images) into 'converted_simplified' instead of full frames.
    :return: None
    """
    output_root_name = "converted_simplified" if simplified else "converted"
    split_path = root_path.joinpath(split)
    split_cache_path = root_path.joinpath(output_root_name).joinpath(split)

    split_path.mkdir(parents=True, exist_ok=True)
    split_cache_path.mkdir(parents=True, exist_ok=True)

    # Cache sequence lengths
    sequence_paths = sorted(list(split_path.iterdir()))
    seq_lens = _cache_seq_lens(sequence_paths, split_cache_path)

    # Generate the new dataset
    frame_count = 0
    for seq_idx, seq_len in tqdm.tqdm(enumerate(seq_lens), total=len(seq_lens)):
        seq_path = sequence_paths[seq_idx]
        seq = tf.data.TFRecordDataset(seq_path, compression_type="")
        for data in seq:
            frame_path = split_cache_path.joinpath(f"{frame_count}.pkl")
            frame_count += 1
            if not frame_path.exists():
                obj = _load_frame(data, simplified=simplified)
                with open(frame_path, "wb") as f:
                    pickle.dump(obj, f)


def _cache_seq_lens(sequence_paths, split_cache_path):
    seq_len_cache_path = split_cache_path.joinpath("len.pkl")
    if seq_len_cache_path.exists():
        with open(seq_len_cache_path, "rb") as f:
            seq_lens = pickle.load(f)
    else:
        print("Computing sequence lengths, might take a while")
        seq_lens = list(tqdm.tqdm((_get_size(s) for s in sequence_paths), total=len(sequence_paths)))
        with open(seq_len_cache_path, "wb") as f:
            pickle.dump(seq_lens, f)
    return seq_lens


def _get_size(s: pathlib.Path) -> int:
    return sum(1 for _ in tf.data.TFRecordDataset(s, compression_type=""))


def _load_frame(data, simplified: bool):
    """
    Load a single frame from TFRecord data.
    :param data: TFRecord data
    :param simplified: If True, return a SimplifiedFrame (no images), else return full Frame.
    :return: Frame or SimplifiedFrame
    """
    frame = open_dataset.Frame()
    frame.ParseFromString(bytearray(data.numpy()))
    converted_frame = dataset_proto.from_data(Frame, frame)

    # Generate point cloud
    (
        range_images,
        camera_projections,
        _,
        range_image_top_pose,
    ) = frame_utils.parse_range_image_and_camera_projection(frame)
    points, _ = frame_utils.convert_range_image_to_point_cloud(
        frame, range_images, camera_projections, range_image_top_pose
    )
    converted_frame.points = points

    if not simplified:
        # Return full frame (images, lasers, labels). Point cloud generation skipped for speed.
        return converted_frame

    # Simplified path: compute point cloud and build SimplifiedFrame (no images stored)
    simple_frame = SimplifiedFrame(
        converted_frame.context,
        converted_frame.timestamp_micros,
        converted_frame.pose,
        converted_frame.laser_labels,
        converted_frame.no_label_zones,
        converted_frame.points,
    )
    return simple_frame


def main():
    SPLITS = ["training", "validation", "testing"]

    parser = argparse.ArgumentParser(
        prog="Convert Waymo",
        description="Convert the Waymo Open Dataset to remove all dependencies to Tensorflow",
    )
    parser.add_argument(
        "-d",
        "--dataset",
        type=str,
        required=True,
        help="Path to the Waymo Open Dataset",
    )
    parser.add_argument(
        "-s",
        "--splits",
        type=str,
        choices=SPLITS,
        nargs="+",
        default=SPLITS,
        help="Specify the splits you want to process",
    )
    parser.add_argument(
        "--simplified",
        action="store_true",
        help="Store simplified frames (no images) into 'converted_simplified' instead of full frames.",
    )

    args = parser.parse_args()
    dataset_path = pathlib.Path(args.dataset).expanduser()
    splits = args.splits
    simplified = args.simplified

    for split in splits:
        mode = "simplified" if simplified else "full"
        print(f"Processing {split} in {mode} mode...")
        generate_cache(dataset_path, split, simplified=simplified)


if __name__ == "__main__":
    main()
